{ Demo program for ChroMathsome - Linear Algebra.

  Example in the field of Statistics.

  In this demonstration program we will use matrices to calculate
  the coefficients of a curve that fits pairs of data points
  aka. linear/polynomial regression.

  As a more challenging example, we create an algorithm to find the coefficients (a1, ..., an)
  of a function
  f(x) = a1 * f1 + a2 * f2 + a3 * f3 + ... + an * fn, where f1, f2, ..., fn are functions itself
  i.e. f(x) is a linear combination of functions, that fits pairs of data sets.

  Note that, since ChroMathsome has its own pre-defined exceptions, you do not need to
  write supersofisticated exception handling routines.
  The only exceptions written here are to make sure the pointers (matrices and vectors)
  get freed, even when an exception has occured (a try ... finally block).

  Since ChroMathsome - Linear Algebra automatically updates the dimensions of
  matrices and vectors (if needed), you can give most of the matrices/vectors in the
  initialization part (InitRealMatrix, ...) a dummy value.
  You are, however, encouraged to enter the correct values (if you know them at forehand)
  since this will make your code much more readable!

  Read the help file for information about the implemented functions
  and their limitations/possibilities!

  Author: John Merrit (john.merrit@net4all.be)
  Release date: June 30, 1998
  Reader may do with this code whatever he/she wants.

  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

  ChroMathsome - Linear Algebra 1996-98 John Merrit
  All Rights Reserved

  Brought to you by the letter "S".              }
unit Stats;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  LinAlg, LinAlgP, Menus, StdCtrls;

type
  TForm1 = class(TForm)
    MainMenu1: TMainMenu;
    Close1: TMenuItem;
    CurveFit1: TMenuItem;
    StraightLine1: TMenuItem;
    Polynom1: TMenuItem;
    Memo1: TMemo;
    LinCombinationofFunctions1: TMenuItem;
    procedure StraightLine1Click(Sender: TObject);
    procedure Polynom1Click(Sender: TObject);
    procedure Close1Click(Sender: TObject);
    procedure LinCombinationofFunctions1Click(Sender: TObject);
  private
    { Private declarations }
    procedure Write_RealVector(V: RealVector);
    procedure Write_RealMatrix(M: RealMatrix);

  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}


{ Linear regression.
  Fit the points (0, 1), (1, 3), (2, 4), (3, 4) into a straight line.
  The least squares solution Y = M * V.
  V gives the coefficients a and b for the line Y = a*X + b.
  where V is given by
           Transpose(M) * Y
       V = ----------------
           Transpose(M) * M
                                                                         }
procedure TForm1.StraightLine1Click(Sender: TObject);
var M, Y, V, M1: RealMatrix;
begin
     InitRealMatrix(M, 4, 2);
     InitRealMatrix(Y, 4, 1);
     InitRealMatrix(V, 2, 1);
     InitRealMatrix(M1, 1, 1);
     try
     Y.Matrix^[1]^[1]:= 1;
     Y.Matrix^[2]^[1]:= 3;
     Y.Matrix^[3]^[1]:= 4;
     Y.Matrix^[4]^[1]:= 4;

     M.Matrix^[1]^[1]:= 0;
     M.Matrix^[1]^[2]:= 1;
     M.Matrix^[2]^[1]:= 1;
     M.Matrix^[2]^[2]:= 1;
     M.Matrix^[3]^[1]:= 2;
     M.Matrix^[3]^[2]:= 1;
     M.Matrix^[4]^[1]:= 3;
     M.Matrix^[4]^[2]:= 1;

     { Find matrix V = [Transpose(M)*Y]/[Transpose(M)*M]. }
     Copy_RealMatrix(M1, M);
     RealMatrix_Transpose(M1);
     RealMatrix_Multiplication(V, M1, M);  { Transpose(M)*M }
     Copy_RealMatrix(M, V);
     RealMatrix_Inverse(M); { A/B = A*Inverse(B) }
     RealMatrix_Multiplication(V, M1, Y);  { Transpose(M)*Y }
     RealMatrix_Multiplication(M1, M, V);

     Memo1.Lines.Add('Coefficients of straight line:');
     Memo1.Lines.Add('Coefficient a :' + FloatToStr(M1.Matrix^[1]^[1]));
     Memo1.Lines.Add('Coefficient b :' + FloatToStr(M1.Matrix^[2]^[1]));
     Memo1.Lines.Add('-> straight line Y = a*X + b');

     finally
     DestroyRealMatrix(M);
     DestroyRealMatrix(Y);
     DestroyRealMatrix(V);
     DestroyRealMatrix(M1);
     end;
end;

{ Polynomial regression.
  According to Newton's 2nd Law of Motion, a body near the earth's surface falls vertically downward
  according to the equation: y = y0 + v0*t + 1/2*g*t (1)
  where
  y = vertical displacement at time t
  y0 = initial vertical displacement at time t0 = 0
  v0 = initial velocity at time t0 = 0
  g = Newton's constant of acceleration of gravity near the earth's surface.

  We wish to evaluate g experimentally.
  A weight is released with unknown initial displacement and velocity.

  At a fixed time interval the distance fallen from a fixed reference point is measured.
  The following results are obtained:
  at times t = 0.1, 0.2, ..., 0.5 seconds the weight has fallen
  y = -0.055, 0.094, 0.314, 0.756 and 1.138 meters, resp., from the reference point.
  Calculate g using this data.

  Solution: fit the quadratic curve y = a + b*X + c*X^2 to the five data points.
            The least squares solution is given by Y = M * V
            where
            Y = [ y1 y2 y3 y4 y5 ]
            M = [ 1 t1 t1^2 ]
                [ 1 t2 t2^2 ]
                [ ... ... . ]
                [ 1 t5 t5^2 ]
            V = [ a b c ]
            
       where V is given by
           Transpose(M) * Y
       V = ----------------
           Transpose(M) * M

  To find g, compare the resulting equation of the parabole with equation (1).
  (for those who do not know this: g ~= 9.81)                                              }
procedure TForm1.Polynom1Click(Sender: TObject);
var M, Y, V, M1: RealMatrix;
begin
     InitRealMatrix(M, 5, 3);
     InitRealMatrix(Y, 5, 1);
     InitRealMatrix(V, 3, 1);
     InitRealMatrix(M1, 1, 1);
     try
     Y.Matrix^[1]^[1]:= -0.055;
     Y.Matrix^[2]^[1]:= 0.094;
     Y.Matrix^[3]^[1]:= 0.314;
     Y.Matrix^[4]^[1]:= 0.756;
     Y.Matrix^[5]^[1]:= 1.138;

     M.Matrix^[1]^[1]:= 1;
     M.Matrix^[1]^[2]:= 0.1;
     M.Matrix^[1]^[3]:= 0.1*0.1;
     M.Matrix^[2]^[1]:= 1;
     M.Matrix^[2]^[2]:= 0.2;
     M.Matrix^[2]^[3]:= 0.2*0.2;
     M.Matrix^[3]^[1]:= 1;
     M.Matrix^[3]^[2]:= 0.3;
     M.Matrix^[3]^[3]:= 0.3*0.3;
     M.Matrix^[4]^[1]:= 1;
     M.Matrix^[4]^[2]:= 0.4;
     M.Matrix^[4]^[3]:= 0.4*0.4;
     M.Matrix^[5]^[1]:= 1;
     M.Matrix^[5]^[2]:= 0.5;
     M.Matrix^[5]^[3]:= 0.5*0.5;

     { Find matrix V = [Transpose(M)*Y]/[Transpose(M)*M]. }
     Copy_RealMatrix(M1, M);
     RealMatrix_Transpose(M1);
     RealMatrix_Multiplication(V, M1, M);  { Transpose(M)*M }
     Copy_RealMatrix(M, V);
     RealMatrix_Inverse(M); { A/B = A*Inverse(B) }
     RealMatrix_Multiplication(V, M1, Y);  { Transpose(M)*Y }
     RealMatrix_Multiplication(M1, M, V);

     Memo1.Lines.Add('Coefficients of polynom:');
     Memo1.Lines.Add('Coefficient a: ' + FloatToStr(M1.Matrix^[1]^[1]));
     Memo1.Lines.Add('Coefficient b: ' + FloatToStr(M1.Matrix^[2]^[1]));
     Memo1.Lines.Add('Coefficient c: ' + FloatToStr(M1.Matrix^[3]^[1]));
     Memo1.Lines.Add('-> polynom Y = a + b*X + c*X^2');
     Memo1.Lines.Add('The experimental value for g = ' + FloatToStr(2*M1.Matrix^[3]^[1]));

     finally
     DestroyRealMatrix(M);
     DestroyRealMatrix(Y);
     DestroyRealMatrix(V);
     DestroyRealMatrix(M1);
     end;
end;

procedure TForm1.Close1Click(Sender: TObject);
begin
     Close;
end;

{ Write a realvector to the screen, creating a nice readable look. }
procedure TForm1.Write_RealVector(V: RealVector);
var a:integer;
    s,ss,st:string;
begin
    Memo1.Enabled:= true;
    st:= '';
    ss:= '[ ';
    for a:= 1 to V.Dimension do
    begin
    if abs(V.Vector^[a]) < 1E-18 then
    begin
    V.Vector^[a]:= 0;
    s:= IntToStr(round(V.Vector^[a]));
    end
    else
    s:= FloatToStrF(V.Vector^[a], ffexponent,5,2);
    st:= st + s + ' ';
    end;
    ss:= ss + st + ']';
    Memo1.Lines.add(ss);
end;

{ Write a real matrix to the screen, creating a nice readable look. }
procedure TForm1.Write_RealMatrix(M: RealMatrix);
var a,b:integer;
    matt,mats,mattt:string;
begin
    memo1.Enabled:= true;
    mattt:= '';
    for a:= 1  to M.Row do
    begin
    mats:= '[ ';
    for b:= 1 to M.Col do
    begin
    if abs(M.Matrix^[a]^[b]) < 1E-17 then
    begin
    M.Matrix^[a]^[b]:= 0;
    matt:= '0';
    end
    else
    matt:= FloatToStrF(M.Matrix^[a]^[b], ffexponent,8,2);
    mattt:= mattt + matt + ' ';
    end;
    mats:= mats + mattt + ']';
    mattt:= '';
    memo1.Lines.add(mats);
    end;
end;

{ Calculate the sum of the elements of a realvector. }
function Sum(V: RealVector): extended;
var i: integer;
    x: extended;
begin
     x:= 0;
     for i:= 1 to V.Dimension do
     x:= x + V.Vector^[i];
     Sum:= x;
end;

{ Calculate the sum of squares of the elements of a realvector. }
function SumOfSquares(V: RealVector): extended;
var i: integer;
    x: extended;
begin
     x:= 0;
     for i:= 1 to V.Dimension do
     x:= x + sqr(V.Vector^[i]);
     SumOfsquares:= x;
end;

{ x^pwr; pwr = integer }
function PowerX(x: extended; pwr:cardinal):extended;
var k: cardinal;
    res: extended;
begin
     res:= 1;
     if pwr = 0 then res:= 1
     else if pwr = 1 then res:= x
     else
     begin
     for k:= 1 to pwr do
     res:= res*x;
     end;
     PowerX:= res;
 end;


{ sum of x^n * y^m }
function SumOfPowerXPowerY(x, y: RealVector; xpwr, ypwr: cardinal):extended;
var i: integer;
    res: extended;
begin
     res:= 0;
     for i:= 1 to x.Dimension do
     res:= res + PowerX(x.Vector^[i], xpwr)*PowerX(y.Vector^[i], ypwr);
     SumOfPowerXPowerY:= res;
end;


{ More challenging example.
  Fit data points X = [ 0.1 0.4 0.5 0.7 0.7 0.9 ] and
  Y = [ 0.61 0.92 0.99 1.52 1.47 2.03 ] into a curve f(x) = a1 + a2*x + a3*sin(x) + a4*exp(x).
  Function f(x) is a linear combination of functions.
  Note: we assume the reader is familiar with the method of least squares.

  This example shows how we can implement a complete "array" of new statistics/regression functions.
  This example can be extended for any kind of linear regression where we wish to find
  regression coefficients for any linear combination of functions.
  Notice that the use of the mathematical function parser from ChroMathsome - Linear Algebra
  indeed enables us to use any mathematical function we want (try writing this scheme without
  matrices/vectors and a mathematical function parser ...)
  The only thing we need to do is populate a matrix and a vector and finally solve a system
  ... et voila les coefficients sont ... ;-) }
procedure TForm1.LinCombinationofFunctions1Click(Sender: TObject);
var M, M1: RealMatrix;
    X, X1, Y, Y1: RealVector;
    det, cf: extended;
    i: integer;
begin
     InitRealMatrix(M, 1, 1);
     InitRealMatrix(M1, 1, 1);
     InitRealVector(X, 1);
     InitRealVector(X1, 1);
     InitRealVector(Y, 1);
     InitRealVector(Y1, 1);

     try
     { Populate Realvectors X and Y. }
     Redimension_RealVector(X, 6);
     Redimension_RealVector(Y, X.Dimension);

     X.Vector^[1]:= 0.1;
     X.Vector^[2]:= 0.4;
     X.Vector^[3]:= 0.5;
     X.Vector^[4]:= 0.7;
     X.Vector^[5]:= 0.7;
     X.Vector^[6]:= 0.9;

     Y.Vector^[1]:= 0.61;
     Y.Vector^[2]:= 0.92;
     Y.Vector^[3]:= 0.99;
     Y.Vector^[4]:= 1.52;
     Y.Vector^[5]:= 1.47;
     Y.Vector^[6]:= 2.03;

     { Populate Realmatrix M.
       Recall from least squares that the structure, for this function (f(x)), is:
       [[ n              sum(xi)           sum(sin(xi))           sum(exp(xi))
          sum(xi)        sum(xi*xi)        sum(xi*sin(xi))        sum(xi*exp(xi))
          sum(sin(xi))   sum(xi*sin(xi))   sum(sqr(sin(xi)))      sum(sin(xi)*exp(xi))
          sum(exp(xi))   sum(xi*exp(xi))   sum(sin(xi)*exp(xi))   sum(sqr(exp(xi)))     ]]
       with xi the i-th component of Realvector X.
       We will find the matrix elements by using the mathematical function parser of
       ChroMathsome - Realistics.                                                      }
     Redimension_Realmatrix(M, 4, 4);
     M.Matrix^[1]^[1]:= X.Dimension; // # of data points
     M.Matrix^[1]^[2]:= Sum(X);
     M.Matrix^[2]^[1]:= Sum(X);
     M.Matrix^[2]^[2]:= SumOfSquares(X);
     Copy_RealVector(X1, X); // Use X1 to apply functions on so we can use X later on if needed.
     RealVector_IterateVector(X1, 'sin(x)');
     M.Matrix^[3]^[1]:= Sum(X1);
     Copy_RealVector(X1, X);
     RealVector_IterateVector(X1, 'exp(x)');
     M.Matrix^[4]^[1]:= Sum(X1);
     Copy_RealVector(X1, X);
     RealVector_IterateVector(X1, 'x*sin(x)');
     M.Matrix^[3]^[2]:= Sum(X1);
     Copy_RealVector(X1, X);
     RealVector_IterateVector(X1, 'x*exp(x)');
     M.Matrix^[4]^[2]:= Sum(X1);
     M.Matrix^[1]^[3]:= M.Matrix^[3]^[1]; // M is a symmetrical matrix!
     M.Matrix^[2]^[3]:= M.Matrix^[3]^[2];
     Copy_RealVector(X1, X);
     RealVector_IterateVector(X1, 'sqr(sin(x))');
     M.Matrix^[3]^[3]:= Sum(X1);
     Copy_RealVector(X1, X);
     RealVector_IterateVector(X1, 'sin(x)*exp(x)');
     M.Matrix^[4]^[3]:= Sum(X1);

     M.Matrix^[1]^[4]:= M.Matrix^[4]^[1];
     M.Matrix^[2]^[4]:= M.Matrix^[4]^[2];
     M.Matrix^[3]^[4]:= M.Matrix^[4]^[3];
     Copy_RealVector(X1, X);
     RealVector_IterateVector(X1, 'sqr(exp(x))');
     M.Matrix^[4]^[4]:= Sum(X1);

     Memo1.Lines.Add('Realmatrix M: ');
     Write_RealMatrix(M);

     { Populate Realvector Y1 in M * A = Y1; A contains the coefficients we're looking for.  }
     { Again, recall from least squares the elements of Y for this function:
       [ sum(yi)
         sum(xi*yi)
         sum(sin(xi)*yi)
         sum(exp(xi)*yi) ]
       with yi, the i-th component of Realvector Y, xi ... }
     det:= 0;
     Redimension_RealVector(Y1, 4);
     Y1.Vector^[1]:= Sum(Y);
     Y1.Vector^[2]:= SumOfPowerXPowerY(X, Y, 1, 1);
     Copy_RealVector(X1, X);
     RealVector_IterateVector(X1, 'sin(x)');
     for i:= 1 to X1.Dimension do
     det:= det + X1.Vector^[i]*Y.Vector^[i];
     Y1.Vector^[3]:= det;
     Copy_RealVector(X1, X);
     det:= 0;
     RealVector_IterateVector(X1, 'exp(x)');
     for i:= 1 to X1.Dimension do
     det:= det + X1.Vector^[i]*Y.Vector^[i];
     Y1.Vector^[4]:= det;

     Memo1.Lines.Add('Realvector Y1: ');
     Write_RealVector(Y1);

     { Now, all needed matrices and vectors are initialized.
       Next, we need to solve the system M * A = Y1 to find vector A containing the coefficients
       of f(x) = a1 + a2*x + a3*sin(x) + a4*exp(x).
       If we use the sysofequation solver from ChroMathsome - Linear Algebra, A (in M * A = Y1)
       will be a matrix  (see help file).
       Note: in "real life" applications, we must check if the system indeed has solution(s) by
       reading the NumOfSolutions variable of RealSysOfEquations(...) (see help file).
       For this example we ignore all the checking (see also: testing the accuracy, lower (**)).
       (I knew the solution at forehand :-))}
     RealSysOfEquations(M1, M, Y1, i);
     Memo1.Lines.Add('Coefficientmatrix M1: ');
     Write_RealMatrix(M1);
     Memo1.Lines.Add('Resulting function;');
     Memo1.Lines.Add(FloatToStr(M1.Matrix^[1]^[1]) + ' + (' + FloatToStr(M1.Matrix^[1]^[2]) + ')*x + (' +
     FloatToStr(M1.Matrix^[1]^[3]) + ')*sin(x) + (' + FloatToStr(M1.Matrix^[1]^[4]) + ')*exp(x)');

     { Finally - Test accuracy of found coefficients.
       Let's find the deviation in every point to get an idea of accuracy of the fit.
       We assume M1 has only one row (**).}
     for i:= 1 to X.Dimension do
     begin
        det:= M1.Matrix^[1]^[1] + M1.Matrix^[1]^[2]*X.Vector^[i] +
        M1.Matrix^[1]^[3]*sin(X.Vector^[i]) + M1.Matrix^[1]^[4]*exp(X.Vector^[i]);
        cf:= abs(det) - abs(Y.Vector^[i]);
        Memo1.Lines.Add('Deviation in point ' + IntToStr(i) + ': ' + FloatToStr(cf));
     end;

     finally
     DestroyRealMatrix(M);
     DestroyRealMatrix(M1);
     DestroyRealVector(X);
     DestroyRealVector(X1);
     DestroyRealVector(Y);
     DestroyRealVector(Y1);
     end;
end;

end.
